//-----------------------------------------------------------------------------
// File: PlayStreaming.cpp
//
// Desc: DirectSound support file for sample showing how to load a wave file 
//       and play it using a streaming DirectSound buffer.
//
// Copyright (c) 1999 Microsoft Corp. All rights reserved.
//
// Gendert und angepasst fr WinMod von Daniel Becker in 2000
//-----------------------------------------------------------------------------
#include "StdAfx.h"
#include <objbase.h>
#include <initguid.h>
#include <commdlg.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <stdio.h>
#include "WinModFileTypes.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//-----------------------------------------------------------------------------
// Function-prototypes
//-----------------------------------------------------------------------------
void createWaveFormatStruct(WAVEFORMATEX *wfx, int nSamplesPerSec, int nBitsPerSample, int nChannels);
HRESULT CreateStreamingBuffer();
HRESULT UpdateProgress();
HRESULT FillBuffer( BOOL bLooped );
HRESULT WriteToBuffer( BOOL bLooped, VOID* pbBuffer, DWORD dwBufferLength );
HRESULT RestoreBuffers( BOOL bLooped );

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
#define SAFE_DELETE(p)  { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_RELEASE(p) { if(p) { (p)->Release(); (p)=NULL; } }
bool CoInit=false;

//-----------------------------------------------------------------------------
// Name: createWaveFormatStruct()
// Desc: Create the WaveFormatStruct...
//-----------------------------------------------------------------------------
void createWaveFormatStruct(WAVEFORMATEX *wfx, int nSamplesPerSec, int nBitsPerSample, int nChannels)
{
    ZeroMemory( wfx, sizeof(WAVEFORMATEX) ); 
    wfx->wFormatTag      = WAVE_FORMAT_PCM;
    wfx->nChannels       = nChannels;
    wfx->nSamplesPerSec  = nSamplesPerSec;
    wfx->wBitsPerSample  = nBitsPerSample;
    wfx->nBlockAlign     = (wfx->wBitsPerSample >> 3) * wfx->nChannels;
    wfx->nAvgBytesPerSec = wfx->nBlockAlign * wfx->nSamplesPerSec;
}

//-----------------------------------------------------------------------------
// Name: InitDirectSound()
// Desc: Initilizes DirectSound
//-----------------------------------------------------------------------------
HRESULT InitDirectSound( HWND hDlg, int nSamplesPerSec, int nBitsPerSample, int nChannels, DWORD WaveCallBack)
{
    WAVEFORMATEX wfx;
	createWaveFormatStruct(&wfx, nSamplesPerSec, nBitsPerSample, nChannels);
    
   	if (useWaveOut)
	{
		MMRESULT result = waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfx, WaveCallBack, 0, CALLBACK_FUNCTION);
		if (result != MMSYSERR_NOERROR) return S_FALSE;
		MMRESULT resultReset = waveOutReset( hWaveOut );
		if (resultReset != MMSYSERR_NOERROR) return S_FALSE;
		return S_OK;
	}
	else
	{
		HRESULT hr;
		LPDIRECTSOUNDBUFFER pDSBPrimary = NULL;

		// Initialize COM
		if( !(hr = CoInitialize(NULL)) )
			CoInit=true;

		// Create IDirectSound using the primary sound device
		if( FAILED( hr = DirectSoundCreate( NULL, &g_pDS, NULL ) ) )
			return hr;

		// Set coop level to DSSCL_PRIORITY
		if( FAILED( hr = g_pDS->SetCooperativeLevel( hDlg, DSSCL_PRIORITY ) ) )
			return hr;

		// Get the primary buffer 
		DSBUFFERDESC        dsbd;
		ZeroMemory( &dsbd, sizeof(DSBUFFERDESC) );
		dsbd.dwSize        = sizeof(DSBUFFERDESC);
		dsbd.dwFlags       = DSBCAPS_PRIMARYBUFFER;
		dsbd.dwBufferBytes = 0;
		dsbd.lpwfxFormat   = NULL;
       
		if( FAILED( hr = g_pDS->CreateSoundBuffer( &dsbd, &pDSBPrimary, NULL ) ) )
			return hr;

		if( FAILED( hr = pDSBPrimary->SetFormat(&wfx) ) )
			return hr;

		SAFE_RELEASE( pDSBPrimary );

		return S_OK;
	}
}

//-----------------------------------------------------------------------------
// Name: FreeDirectSound()
// Desc: Releases DirectSound 
//-----------------------------------------------------------------------------
HRESULT FreeDirectSound()
{
    if (useWaveOut && hWaveOut)
	{
		// Wir mssen weiterlaufen, um den aktuellen Puffer auf jeden Fall zuende
		// zu spielen...
		MMRESULT resultRestart = waveOutRestart( hWaveOut );
		// Buffer alle ungltig machen... Der Callback MUSS deaktiviert worden sein
		// damit nicht weitere Daten an den Treiber gesendet werden...
		for(int i = 0; i < NUM_PLAY_NOTIFICATIONS; i++)
		{
			if(blocks[i].dwFlags & WHDR_PREPARED)
			{
				while(waveOutUnprepareHeader(hWaveOut, &blocks[i], sizeof(WAVEHDR)) == WAVERR_STILLPLAYING) Sleep(0);
			}
		}
		// Dann geht der Reset
		MMRESULT resultReset = waveOutReset( hWaveOut );
		// und der Close
		MMRESULT resultClose = waveOutClose( hWaveOut );
		// Da nun den Speicher keiner mehr braucht....
		HeapFree(GetProcessHeap(), 0, blocks);
		hWaveOut = NULL;
	}
	else
	{
		// Release DirectSound interfaces
		SAFE_RELEASE( lpdsNotify );
		SAFE_RELEASE( g_pDSBuffer );
		SAFE_RELEASE( g_pDS ); 

		// Release COM
		if (CoInit)
		{
			CoUninitialize();
			CoInit=false;
		}
	}
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: CalcBufferLen()
// Desc: Calculates the bufferlength for !one! buffer.
//       The total buffer is this result * NUM_PLAY_NOTIFICATIONS
//-----------------------------------------------------------------------------

DWORD _CalcBufferLen(WAVEFORMATEX &wfx)
{
	DWORD bufferSize = ((DWORD)wfx.nAvgBytesPerSec) * ((DWORD)PlayerBufferLenMillisec)/((DWORD)NUM_PLAY_NOTIFICATIONS)/1000L;
	// Buffer mu in den Sections gro genug sein, um alle Samples aufnehmen zu knnen.
	while (bufferSize % ((DWORD)NUM_PLAY_NOTIFICATIONS)) bufferSize++;
	while (bufferSize % ((DWORD)wfx.nBlockAlign)) bufferSize++;

	return bufferSize;
}

//-----------------------------------------------------------------------------
// Name: CreateStreamingBuffer()
// Desc: Creates a streaming buffer, and the notification events to handle
//       filling it as sound is played
//-----------------------------------------------------------------------------
HRESULT _CreateStreamingBuffer(int nSamplesPerSec, int nBitsPerSample, int nChannels, bool Notify)
{
    HRESULT hr; 
	int i;

    // Set up the direct sound buffer, and only request the flags needed
    // since each requires some overhead and limits if the buffer can 
    // be hardware accelerated
    WAVEFORMATEX wfx;
	createWaveFormatStruct(&wfx, nSamplesPerSec, nBitsPerSample, nChannels);

	DWORD bufferSize = _CalcBufferLen(wfx);
	DWORD totalBufferSize = bufferSize * NUM_PLAY_NOTIFICATIONS;

 	ZeroMemory( &dsbd, sizeof(DSBUFFERDESC) );
    dsbd.dwSize        = sizeof(DSBUFFERDESC);
    dsbd.dwFlags       = DSBCAPS_CTRLPOSITIONNOTIFY | DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_GLOBALFOCUS;
    dsbd.dwBufferBytes = totalBufferSize;
    dsbd.lpwfxFormat   = &wfx;

    // Create a DirectSound buffer
    if( FAILED( hr = g_pDS->CreateSoundBuffer( &dsbd, &g_pDSBuffer, NULL ) ) )
        return hr;

	// Create the Notifications
	if (Notify)
	{
		for (i=0; i<NUM_PLAY_NOTIFICATIONS; i++)
		{
			rghEvent[i] = CreateEvent(NULL, FALSE, FALSE, NULL);
			if (NULL == rghEvent[i]) return FALSE;
		}
	}

	for (i=0; i<NUM_PLAY_NOTIFICATIONS; i++)
	{
	    rgdsbpn[i].dwOffset = (bufferSize * i);
	    if (Notify) rgdsbpn[i].hEventNotify = rghEvent[i];
	}

    // Tell DirectSound when to notify us. the notification will come in the from 
    // of signaled events that are handled in WinMain()
	if (Notify)
	{
		if (FAILED( hr= IDirectSoundBuffer_QueryInterface(g_pDSBuffer, 
						IID_IDirectSoundNotify, (VOID **)&lpdsNotify)))
			return hr; 
 
		if (FAILED( hr=IDirectSoundNotify_SetNotificationPositions(
					   lpdsNotify, NUM_PLAY_NOTIFICATIONS, rgdsbpn)))
		{
			IDirectSoundNotify_Release(lpdsNotify);
			return hr;
		}
	}

    return S_OK;
}

HRESULT _CreateWaveOutBuffer(int nSamplesPerSec, int nBitsPerSample, int nChannels, bool Notify)
{
    WAVEFORMATEX wfx;
	createWaveFormatStruct(&wfx, nSamplesPerSec, nBitsPerSample, nChannels);
	
	DWORD bufferSize = _CalcBufferLen(wfx);
	DWORD totalBufferSize = (bufferSize + sizeof(WAVEHDR)) * NUM_PLAY_NOTIFICATIONS;

	unsigned char *buffer;
    if ((buffer = (unsigned char *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, totalBufferSize)) == NULL)
	{
		return S_FALSE;
    }
    /*
    * and set up the pointers to each bit
    */
    blocks = (WAVEHDR *)buffer;
    buffer += sizeof(WAVEHDR) * NUM_PLAY_NOTIFICATIONS;

    for(int i = 0; i < NUM_PLAY_NOTIFICATIONS; i++)
	{
        blocks[i].dwBufferLength = bufferSize;
        blocks[i].lpData = (char *)buffer;
        buffer += bufferSize;
    }

    return S_OK;
}

HRESULT CreateStreamingBuffer(int nSamplesPerSec, int nBitsPerSample, int nChannels, bool Notify)
{
	if (useWaveOut)
		return _CreateWaveOutBuffer(nSamplesPerSec, nBitsPerSample, nChannels, Notify);
	else
		return _CreateStreamingBuffer(nSamplesPerSec, nBitsPerSample, nChannels, Notify);
}

//-----------------------------------------------------------------------------
// Name: StopBuffer()
// Desc: Stop the DirectSound buffer
//-----------------------------------------------------------------------------
HRESULT StopBuffer() 
{
    if (useWaveOut)
	{
		waveOutPause( hWaveOut );
	}
	else
	{
		if ( g_pDSBuffer ) g_pDSBuffer->Stop();    
	}
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: StartBuffer()
// Desc: Stop the DirectSound buffer
//-----------------------------------------------------------------------------
HRESULT StartBuffer() 
{
    if (useWaveOut)
	{
		waveOutRestart( hWaveOut );
	}
	else
	{
		if ( g_pDSBuffer ) g_pDSBuffer->Play(0, 0, DSBPLAY_LOOPING);
	}
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: PauseBuffer()
// Desc: Pause the DirectSound buffer, if Pause=true, else continue playing
//-----------------------------------------------------------------------------
HRESULT PauseBuffer(bool Pause)
{
	if (Pause) StopBuffer();
	else StartBuffer();
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RetrieveActualNotifySegment()
// Desc: Retrieve the actual Segment, in which DirectSound is playing
//-----------------------------------------------------------------------------
int RetrieveActualNotifySegment()
{
	if (useWaveOut)
	{
		return 0;
	}
	else
	{
		int i;
		bool FirstErg,SecondErg;
		unsigned long ActualPos;

		// Playposition ermitteln:
		g_pDSBuffer->GetCurrentPosition(&ActualPos,NULL);

		for (i=0; i<NUM_PLAY_NOTIFICATIONS; i++)
		{
			// Pos zwischen Offset i und i+1. i+1 aber nicht immer gltig!
			if (i<NUM_PLAY_NOTIFICATIONS-1) FirstErg=ActualPos<rgdsbpn[i+1].dwOffset; else FirstErg=true;
			SecondErg=ActualPos>=rgdsbpn[i].dwOffset;

			if (FirstErg && SecondErg) break;
		}
		return(i);
	}
}